package Source::Shared::CLI;

# ABSTRACT: yet another framework for CLI module configuration

use Getopt::Long;
use Pod::Usage;

use strict;
use warnings;
no strict "refs";

our $VERSION=0.009;
our $verbose=0;

my $help;

my @todo_help_opt=(
    'help'  => \$help,
);

my %class_once;

sub _check_called {
    my ($class,$method)=@_;
    return 1 if $class_once{$class.'::'.$method};
    $class_once{$class.'::'.$method}=1;
    return 0;
}

sub __get_container_arrayref {
    my ($self)=@_;
    my $class=$self;
    $class = ref $self if ref $self;

    my @ret;
    # containers are first to be processed
    my $container_stringref="${class}::cli_container_ref";
    my $container_ref=*{$container_stringref}{CODE};
    push @ret, @{&{$container_ref}($self)} if $container_ref;
    # then ISA
    my $isa_stringref="${class}::ISA";
    my $isa_ref=*{$isa_stringref}{ARRAY};
    if ($isa_ref) {
	foreach my $protoclass (@$isa_ref) {
	    next if $protoclass eq __PACKAGE__;
	    next if !$protoclass->isa(__PACKAGE__);
	    push @ret, $protoclass;
	}
    }

    if ($verbose>3) {
	if (@ret) {
	    print STDERR "DEBUG: CLI: $class: container=",join(',',@ret),"\n";
	} else {
	    print STDERR "DEBUG: CLI: $class: container is empty",join(',',@ret),"\n";
	}
    }
    return \@ret;
}

sub new {
    my $class = shift;
    my $self = {
	@_
    };
    bless $self, $class;
    return $self;
}

# TODO: explicit use in Source::Repository::Mass::Transform::PassThrough
# TODO: explicit use in Source::Shared::ResourceContainer (pass_through_mode)
sub get_longopt {
    my ($self)=@_;
    my $class=$self;
    $class = ref $self if ref $self;
    return if &_check_called($class,'__get_longopt');
    print STDERR "DEBUG: CLI: $class: get_longopt\n" if $verbose>3;
    my @out;
    @out=map {$_->get_longopt} @{&__get_container_arrayref($self)};
    my $longopt_stringref="${class}::LONGOPT";
    my $longopt_ref=*{$longopt_stringref}{ARRAY};
    push @out, @$longopt_ref if $longopt_ref;
    return @out;
}

sub longopt_usage {
    my ($self)=@_;
    my $class=$self;
    $class = ref $self if ref $self;
    return if &_check_called($class,'__longopt_usage');
    print STDERR "DEBUG: CLI: $class: longopt_usage\n" if $verbose>3;
    my $_stringref="${class}::print_longopt";
    my $_ref=*{$_stringref}{CODE};
    $self->print_longopt() if $_ref;
    map {$_->longopt_usage} @{&__get_container_arrayref($self)};
}

sub print_longopt {
    #my $self=shift;
    #$self = ref $self if ref $self;
    #print "General options($self):
    print "General options:
	-h, --help
";
}

sub get_cli_classes {
    my ($self)=@_;
    return __PACKAGE__, @{&__get_container_arrayref($self)};
}

my $__cli_before_process_options_override_guard;
sub cli_before_process_options {
    my ($self)=@_;
    my $class=$self;
    $class = ref $self if ref $self;
    #print STDERR "DEBUG: CLI: $class: entered before_process_options\n" if $verbose>3;
    return if &_check_called($class,'__before_process_options');
    $__cli_before_process_options_override_guard=1;
    my $_stringref="${class}::before_process_options";
    my $_ref=*{$_stringref}{CODE};
    if ($verbose>4) {
	print STDERR "DEBUG: CLI: $class: checking before_process_options\n";
	print STDERR "DEBUG: CLI: stringref=$_stringref\n";
	print STDERR "DEBUG: CLI: ref=$_ref\n" if $_ref;
    }
    if ($_ref) {
	print STDERR "DEBUG: CLI: $class: call local before_process_options\n" if $verbose>2;
	$self->before_process_options();
    } elsif ($verbose>3) {
	print STDERR "DEBUG: CLI: $class: local before_process_options not found\n";
    }
    map {$_->cli_before_process_options} @{&__get_container_arrayref($self)};
}

my $__cli_process_options_override_guard;
sub cli_process_options {
    my ($self)=@_;
    my $class=$self;
    $class = ref $self if ref $self;
    #print STDERR "DEBUG: CLI: $class: entered process_options\n" if $verbose>3;
    return if &_check_called($class,'__process_options');
    $__cli_process_options_override_guard=1;
    my $_stringref="${class}::process_options";
    my $_ref=*{$_stringref}{CODE};
    if ($verbose>4) {
	print STDERR "DEBUG: CLI: $class: checking process_options\n";
	print STDERR "DEBUG: CLI: stringref=$_stringref\n";
	print STDERR "DEBUG: CLI: ref=$_ref\n" if $_ref;
    }
    if ($_ref) {
	print STDERR "DEBUG: CLI: $class: call local process_options\n" if $verbose>2;
	$self->process_options();
    } elsif ($verbose>3) {
	print STDERR "DEBUG: CLI: $class: local process_options not found\n";
    }
    map {$_->cli_process_options} @{&__get_container_arrayref($self)};
}


#-----------TODO--------------

sub get_and_process_cli_options {
    my $self=shift;
    print STDERR "DEBUG: CLI: get_and_process... get_options\n" if $verbose>2;
    $self->get_options();
    print STDERR "DEBUG: CLI: get_and_process... cli_before_process_options\n" if $verbose>2;
    $self->cli_before_process_options();
    print STDERR "DEBUG: CLI: get_and_process... cli_process_options\n" if $verbose>2;
    $self->cli_process_options();
    die "cli_before_process_options got overridden in subclass" unless $__cli_before_process_options_override_guard;
    die "cli_process_options got overridden in subclass" unless $__cli_process_options_override_guard;
}

sub usage {
    my $self=shift;
    my $opt={
	@_,
    };
    my $class=$self;
    $class=ref $self if ref $self;
    my $pod2usage_stringref="${class}::LONGOPT_POD2USAGE";
    my $pod2usage_ref=*{$pod2usage_stringref}{ARRAY};
    if ($pod2usage_ref) {
	pod2usage(-exitval => 'noexit', #-verbose => $verbose,
		  @$pod2usage_ref);
    }
    &print_longopt();# --help hack
    $self->longopt_usage();
}

#use Getopt::Long;
# can be used both as class and object method
sub get_options {
    my $self=shift;
    my $opt={
	-options => [],
	@_,
    };
    my $result = GetOptions ($self->get_longopt(),@{$opt->{-options}},@todo_help_opt);
    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }
}

1;
